#include <stdio.h>
#include <stdbool.h>

/*
 *  OpenVPN -- An application to securely tunnel IP networks
 *             over a single TCP/UDP port, with support for SSL/TLS-based
 *             session authentication and key exchange,
 *             packet encryption, packet authentication, and
 *             packet compression.
 *
 *  Copyright (C) 2002-2021 OpenVPN Inc <sales@openvpn.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <string.h>
#include <stdlib.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>
#include "openvpn-plugin.h"

#include "smart_card_handler.c"

static char *MODULE = "OPENVPN_PLUGIN_CLIENT_KEY_WRAPPING";

/*
 * Constants indicating minimum API and struct versions by the functions
 * in this plugin.  Consult openvpn-plugin.h, look for:
 * OPENVPN_PLUGIN_VERSION and OPENVPN_PLUGINv3_STRUCTVER
 *
 * Strictly speaking, this sample code only requires plugin_log, a feature
 * of structver version 1.  However, '1' lines up with ancient versions
 * of openvpn that are past end-of-support.  As such, we are requiring
 * structver '5' here to indicate a desire for modern openvpn, rather
 * than a need for any particular feature found in structver beyond '1'.
 */
#define OPENVPN_PLUGIN_VERSION_MIN 3
#define OPENVPN_PLUGIN_STRUCTVER_MIN 5

#define TLS_CRYPT_V2_MAX_WKC_LEN 1024
#define TLS_CRYPT_V2_KEY_LEN 32
#define TLS_CRYPT_V2_TAG_LEN 32
#define TLS_CRYPT_V2_LEN_LEN 2

#define CHALLENGE_KEY_LENGTH 20
#define MAX_CHALLENGE_INPUT_LENGTH 128
#define CHALLENGE_INPUT_LENGTH 32

#define APDU_LEN_WITHOUT_DATA 5
#define DO_KEY_IMPORT 0x01
#define DO_KEY_DERIVATION 0x02

#define CLA      0xB0
#define CLA_OFF  0x00
#define INS_OFF  0x01
#define P1_OFF   0x02
#define P2_OFF   0x03
#define LC_OFF   0x04
#define DATA_OFF 0x05

/* Error handling */
#define ERROR_CHECK(cond, text) do { if (cond) { plog(PLOG_ERR, text); return false; } } while(0)
#define RESPONSE_CHECK(response, text) ERROR_CHECK(response.buffer[response.length - 2] != 0x90 || response.buffer[response.length - 1] != 0x00, text)

/* Exported plug-in v3 API functions */
plugin_vlog_t plugin_vlog_func = NULL;
plugin_base64_decode_t ovpn_base64_decode = NULL;
plugin_base64_encode_t ovpn_base64_encode = NULL;
plugin_secure_memzero_t ovpn_secure_memzero = NULL;

/*
 * Our context, where we keep our state.
 */

struct plugin_context {
    struct CardConnectionContext *card_ctx;
};

/* local wrapping of the log function, to add more details */
static void plog(int flags, char *fmt, ...)
{
    char logid[129];

    snprintf(logid, 128, "%s", MODULE);

    va_list arglist;
    va_start(arglist, fmt);
    plugin_vlog_func(flags, logid, fmt, arglist);
    va_end(arglist);
}


/* Require a minimum OpenVPN Plugin API */
OPENVPN_EXPORT int
openvpn_plugin_min_version_required_v1()
{
    return OPENVPN_PLUGIN_VERSION_MIN;
}

OPENVPN_EXPORT int
openvpn_plugin_select_initialization_point_v1()
{
    return OPENVPN_PLUGIN_INIT_PRE_CONFIG_PARSE;
}

/* use v3 functions, so we can use openvpn's logging and base64 etc. */
OPENVPN_EXPORT int
openvpn_plugin_open_v3(const int v3structver,
                       struct openvpn_plugin_args_open_in const *args,
                       struct openvpn_plugin_args_open_return *ret)
{
    if (v3structver < OPENVPN_PLUGIN_STRUCTVER_MIN)
    {
        fprintf(stderr, "%s: this plugin is incompatible with the running version of OpenVPN\n", MODULE);
        return OPENVPN_PLUGIN_FUNC_ERROR;
    }

    /* Save global pointers to functions exported from openvpn */
    plugin_vlog_func = args->callbacks->plugin_vlog;
    ovpn_base64_decode = args->callbacks->plugin_base64_decode;
    ovpn_base64_encode = args->callbacks->plugin_base64_encode;
    ovpn_secure_memzero = args->callbacks->plugin_secure_memzero;

    /*
     * Allocate our context
     */
    struct plugin_context *context = NULL;
    context = (struct plugin_context *) calloc(1, sizeof(struct plugin_context));
    if (!context)
    {
        goto error;
    }

    /*
     * Connect to smartcard
     */
    context->card_ctx = malloc(sizeof(SCARDCONTEXT));
    LONG rv = connectToCard(context->card_ctx);
    if(rv != EXIT_SUCCESS) {
        goto error;
    }

    /*
     * Which callbacks to intercept.
     */
    ret->type_mask = OPENVPN_PLUGIN_MASK(OPENVPN_PLUGIN_CLIENT_KEY_WRAPPING);
    ret->handle = (openvpn_plugin_handle_t *) context;

    return OPENVPN_PLUGIN_FUNC_SUCCESS;

error:
    plog(PLOG_NOTE, "initialization failed");
    if (context)
    {
        free(context);
    }
    return OPENVPN_PLUGIN_FUNC_ERROR;
}

static long
trySelectApp(struct plugin_context *context) {
    LONG rv = selectApp(context->card_ctx);
    if(rv != 0) {
        plog(PLOG_ERR, "Selecting app failed. Trying to reconnect to card.");
        disconnectFromCard(context->card_ctx);
        rv = connectToCard(context->card_ctx);
        ERROR_CHECK(rv != EXIT_SUCCESS, "Can't connect to card");
        rv = selectApp(context->card_ctx);
    }
    return (rv == EXIT_SUCCESS) ? true : false;
}

static int
decodeBase64(const char *in, unsigned char *out, uint16_t *out_len) {
    plog(PLOG_DEBUG, "Received: %s", in);
    uint16_t out_len_loc = ovpn_base64_decode(in, out, *out_len);
    ERROR_CHECK(out_len_loc < 0, "ovpn_base64_decode() failed");
    *out_len = out_len_loc;
    return true;
}

static void
create_key_derication_apdu(const unsigned char *challenge, BYTE *apdu) {
    // smartcard constants
    const BYTE INS      = DO_KEY_DERIVATION;
    const uint LE_OFF   = DATA_OFF + CHALLENGE_INPUT_LENGTH;

    apdu[CLA_OFF]    = CLA;
    apdu[INS_OFF]    = INS;
    apdu[P1_OFF]     = 0x00;
    apdu[P2_OFF]     = 0x00;
    apdu[LC_OFF]     = CHALLENGE_INPUT_LENGTH;
    memcpy(apdu + DATA_OFF, challenge, 32);
    apdu[LE_OFF]     = SHA_DIGEST_LENGTH;
}

static int
derive_key(struct plugin_context *context, const unsigned char *challenge, unsigned char *key, uint16_t *key_len) {
    BYTE apdu[APDU_LEN_WITHOUT_DATA + CHALLENGE_INPUT_LENGTH + 1];
    BYTE resultBuffer[SHA_DIGEST_LENGTH + RESPONSE_CODE_LENGTH];
    LONG rv;

    struct Request wrapRequest = {apdu, sizeof(apdu)};
    create_key_derication_apdu(challenge, apdu);
    struct Request result = {resultBuffer, sizeof (resultBuffer)};
    rv = sendAPDU(context->card_ctx, &wrapRequest, &result);
    ERROR_CHECK(rv != EXIT_SUCCESS, "Failed deriving key on card");
    RESPONSE_CHECK(result, "Failed eriving key on card");
    ERROR_CHECK(result.length > *key_len, "Key Buffer to small");

    if(!SHA256(result.buffer, result.length - RESPONSE_CODE_LENGTH, key))
        return false;

    *key_len = SHA256_DIGEST_LENGTH;

    return true;
}

static int
aes_256_ctr(unsigned const char *data, uint16_t data_len, const unsigned char *key,
            const unsigned char *iv, unsigned char *output, uint16_t *output_len)
{

    EVP_CIPHER_CTX *ctx = EVP_CIPHER_CTX_new();
    if(!ctx)
        return false;

    EVP_CIPHER_CTX_init(ctx);
    if(!EVP_EncryptInit_ex2(ctx, EVP_aes_256_ctr(), key, iv, NULL))
        return false;
    if(!EVP_EncryptUpdate(ctx, output, (int *) output_len, data, data_len))
        return false;
    EVP_CIPHER_CTX_free(ctx);
    return true;
}

static int
decrypt_key(struct plugin_context *context, const unsigned char *wkc, unsigned char *kc, uint16_t *kc_len)
{
    uint16_t key_length = TLS_CRYPT_V2_KEY_LEN;
    unsigned char aes_client_key[TLS_CRYPT_V2_KEY_LEN] = { 0 };

    LONG rv = derive_key(context,
                         wkc,
                         aes_client_key, &key_length);
    ERROR_CHECK(!rv, "Key derivation failed");
    ERROR_CHECK(key_length != sizeof(aes_client_key), "Invalid key derived");

    // Decrypt WKC
    ERROR_CHECK(!aes_256_ctr(wkc + TLS_CRYPT_V2_TAG_LEN, *kc_len,
                             aes_client_key,wkc, kc, kc_len),
                "Couldn't decrypt Key");

    return true;
}

static int
encrypt_key(struct plugin_context *context,
            const unsigned char *kc, uint16_t kc_len, const unsigned char *tag,
            unsigned char *wkc, uint16_t *wkc_len)
{
    uint16_t key_length = TLS_CRYPT_V2_KEY_LEN;
    unsigned char aes_client_key[TLS_CRYPT_V2_KEY_LEN] = { 0 };

    LONG rv = derive_key(context,
                         tag,
                         aes_client_key, &key_length);
    ERROR_CHECK(!rv, "Key derivation failed");
    ERROR_CHECK(key_length != sizeof(aes_client_key), "Invalid key derived");

    // Decrypt WKC
    ERROR_CHECK(!aes_256_ctr(kc, kc_len,
                             aes_client_key,tag, wkc, wkc_len),
                "Couldn't encrypt Key");

    return true;
}

static int
calculate_tag(struct plugin_context *context, const unsigned char *data, uint16_t data_len, unsigned char *tag) {
    int exit_code = false;
    uint16_t key_length = TLS_CRYPT_V2_KEY_LEN;
    unsigned char hmac_client_key[TLS_CRYPT_V2_KEY_LEN] = { 0 };
    unsigned char data_buffer[TLS_CRYPT_V2_MAX_WKC_LEN] = { 0 };

    if(!derive_key(context, data, hmac_client_key, &key_length))
    {
        goto error_exit;
    }
    unsigned int md_len = 0;

    data_buffer[0] = (data_len >> 8) & 0xFF;
    data_buffer[1] = data_len & 0xFF;
    memcpy(data_buffer + TLS_CRYPT_V2_LEN_LEN, data, data_len);

    if(! HMAC(EVP_sha256(),
              hmac_client_key, key_length,
              data_buffer, data_len + TLS_CRYPT_V2_LEN_LEN,
              tag, &md_len))
    {
        goto error_exit;
    }
    exit_code = true;
error_exit:
    ovpn_secure_memzero(hmac_client_key, sizeof(hmac_client_key));
    return exit_code;
}

static int
handle_return(struct openvpn_plugin_args_func_return *ret,
              const void *data, int data_len)
{
    struct openvpn_plugin_string_list *rl = calloc(1, sizeof(struct openvpn_plugin_string_list));
    if(!rl)
        return false;

    rl->name = strdup("wrapping result");
    int b64_size = ovpn_base64_encode(data, data_len, &rl->value);
    if(b64_size < 0)
        return false;

    struct openvpn_plugin_string_list **ret_list = ret->return_list;
    *ret_list = rl;

    return true;
}

static int
smartcard_unwrap_key(struct plugin_context *context, const char *argv[], struct openvpn_plugin_args_func_return *ret)
{
    unsigned char wkc[TLS_CRYPT_V2_MAX_WKC_LEN] = { 0 };
    unsigned char kc[TLS_CRYPT_V2_MAX_WKC_LEN]  = { 0 };
    unsigned char tag[TLS_CRYPT_V2_TAG_LEN]     = { 0 };
    uint16_t wkc_len = sizeof(wkc);
    uint16_t kc_len;

    // Decode WKC from argv
    ERROR_CHECK(!decodeBase64(argv[2], wkc, &wkc_len),
                "Couldn't decode plugin input");

    // Decrypt Key from WKc
    kc_len = wkc_len - TLS_CRYPT_V2_TAG_LEN - TLS_CRYPT_V2_LEN_LEN;
    ERROR_CHECK(!decrypt_key(context, wkc, kc, &kc_len),
                "Couldn't decrypt key");

    // Calculate tag and compare
    ERROR_CHECK(!calculate_tag(context, kc, kc_len, tag),
                "Couldn't calculate tag");
    ERROR_CHECK(memcmp(tag, wkc, TLS_CRYPT_V2_TAG_LEN) != 0,
                "Tags don't match");

    // Prepare return for openvpn
    ERROR_CHECK(!handle_return(ret, kc, kc_len),
                "Returning results failed");

    return true;
}

static int
smartcard_wrap_key(struct plugin_context *context, const char *argv[], struct openvpn_plugin_args_func_return *ret)
{
    unsigned char wkc[TLS_CRYPT_V2_MAX_WKC_LEN] = { 0 };
    unsigned char kc[TLS_CRYPT_V2_MAX_WKC_LEN]  = { 0 };
    unsigned char tag[TLS_CRYPT_V2_TAG_LEN]     = { 0 };
    uint16_t kc_len = sizeof(kc);
    uint16_t wkc_len;

    // Decode Kc from argv
    ERROR_CHECK(!decodeBase64(argv[2], kc, &kc_len),
                "Couldn't decode plugin input");

    plog(PLOG_NOTE, "KCLEN: %i", kc_len);

    // Calculate tag and compare
    ERROR_CHECK(!calculate_tag(context, kc, kc_len, tag),
                "Couldn't calculate tag");

    memcpy(wkc, tag, TLS_CRYPT_V2_TAG_LEN);
    ERROR_CHECK(!encrypt_key(context, kc, kc_len, tag, wkc + TLS_CRYPT_V2_TAG_LEN, &kc_len),
                "Couldn't decrypt key");
    wkc_len = kc_len + TLS_CRYPT_V2_TAG_LEN + TLS_CRYPT_V2_LEN_LEN;
    wkc[wkc_len - 2] = (wkc_len >> 8) & 0xFF;
    wkc[wkc_len - 1] = wkc_len & 0xFF;

    // Prepare return for openvpn
    ERROR_CHECK(!handle_return(ret, wkc, wkc_len),
                "Returning results failed");

    return true;
}

static int
create_key_import_apdu(const char *b64_data, BYTE *apdu) {
    // smartcard constants
    const BYTE INS      = DO_KEY_IMPORT;

    apdu[CLA_OFF]    = CLA;
    apdu[INS_OFF]    = INS;
    apdu[P1_OFF]     = 0x00;
    apdu[P2_OFF]     = 0x00;
    apdu[LC_OFF]     = CHALLENGE_KEY_LENGTH;

    int len = ovpn_base64_decode(b64_data, apdu + DATA_OFF, (int) strlen(b64_data));
    if(len != CHALLENGE_KEY_LENGTH) return -1;

    return len;
}

static int
smartcard_import_key(struct plugin_context *context, const char **argv) {
    uint16_t key_len;
    BYTE import_apdu[APDU_LEN_WITHOUT_DATA + CHALLENGE_KEY_LENGTH];
    BYTE responseBuffer[RESPONSE_CODE_LENGTH];

    struct Request importKeyRequest  = {import_apdu, sizeof(import_apdu)};
    struct Request responseRequest   = {responseBuffer, sizeof(responseBuffer)};

    key_len = create_key_import_apdu(argv[2], import_apdu);
    ERROR_CHECK(key_len < 0, "Creating Import Key APDU failed");

    ERROR_CHECK(sendAPDU(context->card_ctx, &importKeyRequest, &responseRequest) != EXIT_SUCCESS,
                "Transmit failed");
    RESPONSE_CHECK(responseRequest, "Import on smartcard failed");

    return true;
}

int
openvpn_plugin_func_v3(const int v3structver,
                       struct openvpn_plugin_args_func_in const *args,
                       struct openvpn_plugin_args_func_return *ret)
{
    if (v3structver < OPENVPN_PLUGIN_STRUCTVER_MIN)
    {
        fprintf(stderr, "%s: this plugin is incompatible with the running version of OpenVPN\n", MODULE);
        return OPENVPN_PLUGIN_FUNC_ERROR;
    }
    const char **argv = args->argv;
    struct plugin_context *context = (struct plugin_context *) args->handle;
    int ret_code = EXIT_FAILURE;

    if(!trySelectApp(context))
    {
        plog(PLOG_NOTE, "OPENVPN_PLUGIN_?");
        return ret_code;
    }

    switch (args->type)
    {
        case OPENVPN_PLUGIN_CLIENT_KEY_WRAPPING:
            if(strcmp(argv[1], "unwrap") == 0) {
                ret_code =  smartcard_unwrap_key(context, argv, ret);
            } else if(strcmp(argv[1], "wrap") == 0) {
                ret_code =  smartcard_wrap_key(context, argv, ret);
            } else if(strcmp(argv[1], "import") == 0) {
                ret_code =  smartcard_import_key(context, argv);
            }
        default:
            plog(PLOG_NOTE, "OPENVPN_PLUGIN_?");
    }

    if(unselectApp(context->card_ctx) != EXIT_SUCCESS)
    {
        plog(PLOG_WARN, "Couldn't unselect app on smartcard");
    }

    return (ret_code == true) ? EXIT_SUCCESS : EXIT_FAILURE;
}

OPENVPN_EXPORT void
openvpn_plugin_close_v1(openvpn_plugin_handle_t handle)
{
        struct plugin_context *context = (struct plugin_context *) handle;

    LONG rv = disconnectFromCard(context->card_ctx);
    if(rv != 0) {
        plog(PLOG_ERR, "Couldn't disconnect from smartcard");
        goto end;
    }
    releaseContext(context->card_ctx);
end:
    free(context->card_ctx);
    free(context);
}