#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <bits/stdint-uintn.h>

#include <winscard.h>

const BYTE SELECT[] = {0x00, 0xA4, 0x04, 0x00};
const BYTE CHOOSE[] = {0x08};
const BYTE UNCHOOSE[] = {0x00};
const BYTE APP_ID[] = {0x0F, 0x0E, 0x0D, 0x0C, 0x0B, 0x0A, 0x09, 0x08};

struct CardConnectionContext {
    SCARDCONTEXT hContext;
    SCARDHANDLE hCard;
    const SCARD_IO_REQUEST *pioSendPci;
};

struct Request {
    BYTE *buffer;
    DWORD length;
};

int connectToCard(struct CardConnectionContext *ctx);
int sendAPDU(struct CardConnectionContext *ctx, struct Request *sendBuffer, struct Request *receiveBuffer);
int disconnectFromCard(struct CardConnectionContext *ctx);
int selectApp(struct CardConnectionContext *ctx);
int unselectApp(struct CardConnectionContext *ctx);


long getReaderList(struct CardConnectionContext *ctx, char ***readers) {
    LONG rv;
    DWORD dwReaders;
    LPSTR mszReaders = NULL;
    int nbReaders;
    char *ptr = NULL;

    /* Retrieve the available readers list. */
    dwReaders = SCARD_AUTOALLOCATE;
    rv = SCardListReaders(ctx->hContext, NULL, (LPSTR)&mszReaders, &dwReaders);
    if(rv != SCARD_S_SUCCESS)
    {
        puts("SCardListReaders");
        return EXIT_FAILURE;
    }

    /* Extract readers from the null separated string and get the total
     * number of readers */
    nbReaders = 0;
    ptr = mszReaders;
    while (*ptr != '\0')
    {
        ptr += strlen(ptr)+1;
        nbReaders++;
    }

    if (nbReaders == 0)
    {
        puts("No readers");
        return EXIT_FAILURE;
    }

    /* allocate the readers table */
    *readers = calloc(nbReaders, sizeof(char *));
    if (readers == NULL)
    {
        puts("malloc(readers table) failed");
        return EXIT_FAILURE;
    }

    /* fill the readers table */
    nbReaders = 0;
    ptr = mszReaders;
    while (*ptr != '\0')
    {
        // printf("%d: %s\n", nbReaders, ptr);
        *readers[nbReaders] = ptr;
        ptr += strlen(ptr)+1;
        nbReaders++;
    }

    return EXIT_SUCCESS;
}

int connectToCard(struct CardConnectionContext *ctx) {
    LONG rv;
    char **readers = NULL;
    DWORD dwActiveProtocol;
    int readerIndex;

    rv = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &ctx->hContext);
    if (rv != SCARD_S_SUCCESS)
    {
        printf("SCardEstablishContext: Cannot Connect to Resource Manager %lX\n", rv);
        return EXIT_FAILURE;
    }

    rv = getReaderList(ctx, &readers);
    if(rv == EXIT_FAILURE) {
        puts("Cannot get list of readers");
        return EXIT_FAILURE;
    }

    readerIndex = 0;

    /* connect to a card */
    dwActiveProtocol = -1;
    rv = SCardConnect(ctx->hContext, readers[readerIndex], SCARD_SHARE_SHARED,
                      SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1, &ctx->hCard, &dwActiveProtocol);
    if(rv != SCARD_S_SUCCESS) {
        puts("SCardConnect");
        return EXIT_FAILURE;
    }

    switch(dwActiveProtocol)
    {
        case SCARD_PROTOCOL_T0:
            ctx->pioSendPci = SCARD_PCI_T0;
            break;
        case SCARD_PROTOCOL_T1:
            ctx->pioSendPci = SCARD_PCI_T1;
            break;
        default:
            puts("Unknown protocol");
            return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

void printRequest(struct Request *req, char title[]) {
    printf("%s", title);
    for (int i=0; i < req->length; i++)
        printf("%02X ", req->buffer[i]);
    printf("\n");
}

int sendAPDU(struct CardConnectionContext *ctx, struct Request *sendBuffer, struct Request *receiveBuffer) {
    SCARD_IO_REQUEST pioRecvPci;
    LONG rv;

    printRequest(sendBuffer, ">> ");
    rv = SCardTransmit(ctx->hCard, ctx->pioSendPci, sendBuffer->buffer, sendBuffer->length,
                       &pioRecvPci, receiveBuffer->buffer, &receiveBuffer->length);

    printRequest(receiveBuffer, "<<  ");
    if(rv != SCARD_S_SUCCESS)
    {
        puts("SCardTransmit");
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

int disconnectFromCard(struct CardConnectionContext *ctx) {
    if(SCardDisconnect(ctx->hCard, SCARD_RESET_CARD) != SCARD_S_SUCCESS)
    {
        puts("SCardDisconnect");
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

int selectApp(struct CardConnectionContext *ctx) {
    const uint apduLength = sizeof(SELECT) + sizeof(CHOOSE) + sizeof(APP_ID);
    BYTE selectAPDU[apduLength];
    memcpy(selectAPDU, SELECT, sizeof(SELECT));
    memcpy(selectAPDU+sizeof(SELECT), CHOOSE, sizeof(CHOOSE));
    memcpy(selectAPDU+sizeof(SELECT)+sizeof(CHOOSE), APP_ID, sizeof(APP_ID));

    struct Request selectRequest = {selectAPDU, apduLength};
    BYTE pbRecvBuffer[2];
    struct Request recvBuffer = {pbRecvBuffer, sizeof(pbRecvBuffer)};
    SCardBeginTransaction(ctx->hCard);
    return sendAPDU(ctx, &selectRequest, &recvBuffer);
}

int unselectApp(struct CardConnectionContext *ctx) {
    const uint apduLength = sizeof(SELECT) + sizeof(UNCHOOSE);
    BYTE unselectAPDU[apduLength];
    memcpy(unselectAPDU, SELECT, sizeof(SELECT));
    memcpy(unselectAPDU + sizeof(SELECT), CHOOSE, sizeof(UNCHOOSE));

    struct Request unselectRequest = {unselectAPDU, apduLength};
    BYTE pbRecvBuffer[2];
    struct Request recvBuffer = {pbRecvBuffer, sizeof(pbRecvBuffer)};
    int status = sendAPDU(ctx, &unselectRequest, &recvBuffer);
    SCardEndTransaction(ctx->hCard, SCARD_LEAVE_CARD);
    return status;
}

void releaseContext(struct CardConnectionContext *ctx) {
    LONG rv = SCardReleaseContext(ctx->hContext);
    if (rv != SCARD_S_SUCCESS)
        printf("SCardReleaseContext: %s (0x%lX)\n", pcsc_stringify_error(rv), rv);
}

int main(int argc, char *argv[])
{
    LONG rv;
    struct CardConnectionContext ctx;
    rv = connectToCard(&ctx);
    if(rv != 0) {
        goto end;
    }

    rv = selectApp(&ctx);
    if(rv != 0) {
        goto end;
    }

    uint8_t importKey[] = {0xB0, 0x01, 0x00, 0x00, 0x14, 0x7A, 0x41, 0xEA, 0xCB, 0xBF, 0xE0, 0x1A, 0x91, 0x9C, 0x2D, 0xBF, 0xC8, 0x3E, 0xE4, 0x5B, 0x1A, 0xD0, 0x67, 0xD0, 0x90};
    struct Request importAESRequest = {importKey, sizeof(importKey)};
    BYTE pbRecvBuffer[2];
    struct Request recvBuffer = {pbRecvBuffer, sizeof(pbRecvBuffer)};
    sendAPDU(&ctx, &importAESRequest, &recvBuffer);

    uint8_t deriveKey[] = {0xB0, 0x02, 0x00, 0x00, 0x20, 0xB4, 0x2C, 0xCC, 0x22, 0x2C, 0x86, 0x58, 0x5E, 0x8F, 0x76, 0x1F, 0x15, 0xBD, 0xFF, 0xCC, 0x77, 0x93, 0xA5, 0x18, 0x58, 0x7A, 0xBB, 0x0D, 0xDF, 0x0C, 0x4A, 0xDB, 0xD4, 0x65, 0x65, 0x1C, 0x07, 0x14};
    struct Request unwrapKeyRequest = {deriveKey, sizeof(deriveKey)};
    BYTE keyBuffer[1024];
    struct Request keyBufferRequest = {keyBuffer, sizeof(keyBuffer)};
    sendAPDU(&ctx, &unwrapKeyRequest, &keyBufferRequest);

    if(keyBufferRequest.buffer[keyBufferRequest.length - 2] != 0x90
       || keyBufferRequest.buffer[keyBufferRequest.length - 1] != 0x00)
    {
        puts("Derive on smartcard failed");
    }

    rv = unselectApp(&ctx);
    if(rv != 0) {
        goto end;
    }

    rv = disconnectFromCard(&ctx);
    if(rv != 0) {
        goto end;
    }
end:
    releaseContext(&ctx);
    return EXIT_SUCCESS;
} /* main */

